"""GraphQL API query functions."""
from typing import Any, Dict, List, Optional, Tuple, Union

import requests
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.timezone import now as tz_now

from linear.models import LinearIssue

PAGE_SIZE = getattr(settings, "LINEAR_API_PAGE_SIZE", 100)

QueryResponseType = Dict[str, Any]
PageInfoType = Dict[str, Union[str, bool]]
PagedIssuesResponseType = Tuple[List[LinearIssue], PageInfoType]


class LinearApiError(Exception):
    pass


def parse_issue_node(issue_data: dict) -> LinearIssue:
    """Parse GraphQL response node into a new LinearIssue object (unsaved)."""
    id = issue_data["id"]
    identifier = issue_data["identifier"]
    project = issue_data["project"] or {"milestone": None}
    milestone = project["milestone"] or {}
    team = issue_data["team"] or {}
    title = issue_data["title"]
    estimate = issue_data["estimate"]
    assignee = issue_data["assignee"] or {}
    state = issue_data["state"]
    created_at = issue_data["createdAt"]
    last_updated_at = issue_data["updatedAt"]
    return LinearIssue(
        id=id,
        identifier=identifier,
        team_name=team.get("name"),
        project_name=project.get("name", ""),
        milestone_name=milestone.get("name", ""),
        title=title,
        estimate=estimate,
        assignee_name=assignee.get("name", ""),
        state=state.get("name", ""),
        created_at=created_at,
        last_updated_at=last_updated_at,
    )


def run_query(query: str, **variables: Any) -> QueryResponseType:
    """Run GraphQL query and return the "data" contents."""
    if not settings.LINEAR_API_KEY:
        raise ImproperlyConfigured("LINEAR_API_KEY setting is missing")

    response = requests.post(
        url="https://api.linear.app/graphql",
        headers={
            "Content-Type": "application/json",
            "Authorization": settings.LINEAR_API_KEY,
        },
        json={
            "query": query,
            "variables": variables,
        },
    )
    payload = response.json()
    if "errors" in payload:
        msg = "; ".join([e["message"] for e in payload["errors"]])
        raise LinearApiError(msg)
    return payload["data"]


def fetch_issues(
    page_size: int = PAGE_SIZE, after: Optional[str] = None, save: bool = False
) -> PagedIssuesResponseType:
    """Fetch a page of issues from the API, without saving them."""
    query = """
      query ($pageSize: Int $after: String) {
        issues(first: $pageSize after: $after) {
          nodes {
            id
            identifier
            team {
              name
            }
            project {
              name
              milestone {
                name
              }
            }
            title
            estimate
            assignee {
              name
            }
            state {
              name
            }
            createdAt
            updatedAt
          }
          pageInfo {
            startCursor
            endCursor
            hasNextPage
          }
        }
      }
    """
    data = run_query(query, after=after, pageSize=page_size)
    page_info = data["issues"]["pageInfo"]
    issues: List[LinearIssue] = []
    for node in data["issues"]["nodes"]:
        issue = parse_issue_node(node)
        issue.last_refreshed_at = tz_now()
        issues.append(issue)
    return issues, page_info


def pull_issues(
    page_size: int = PAGE_SIZE, after: Optional[str] = None, save: bool = False
) -> PagedIssuesResponseType:
    """Fetch a page of issues from the API and save them locally."""
    issues, page_info = fetch_issues(page_size=page_size, after=after)
    for issue in issues:
        issue.save()
    return issues, page_info


def fetch_issue(id: str) -> LinearIssue:
    """Fetch an issue by id from the Linear API, without saving."""
    query = """
      query ($id: String!) {
        issue(id: $id) {
          id
          identifier
          team{name}
          project{name milestone{name}}
          title
          estimate
          assignee{name}
          state{name}
        }
      }
    """
    data = run_query(query, id=id)
    issue = parse_issue_node(data["issue"])
    issue.last_refreshed_at = tz_now()
    return issue


def pull_issue(id: str) -> LinearIssue:
    """Fetch an issue by id from the Linear API and save locally."""
    return fetch_issue(id=id).save()



import json
import logging

from django.contrib import messages
from django.contrib.auth.decorators import login_required, user_passes_test
from django.core.management import call_command
from django.http import HttpRequest, HttpResponse
from django.http.response import HttpResponseRedirect
from django.urls import reverse
from django.utils.timezone import now as tz_now
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods

from .models import LinearIssue

logger = logging.getLogger(__name__)


@login_required
@user_passes_test(lambda u: u.is_staff)
def import_issues(request: HttpRequest) -> HttpResponseRedirect:
    if not request.user.has_perms("linear.use_api"):
        messages.add_message(
            request, messages.ERROR, "You do not permission to use the Linear API"
        )
    else:
        call_command("import_issues")
        messages.add_message(
            request,
            messages.SUCCESS,
            "All Linear issues have been imported successfully",
        )
    return HttpResponseRedirect(reverse("admin:linear_linearissue_changelist"))


@csrf_exempt
@require_http_methods(["POST"])
def webhook(request: HttpRequest) -> HttpResponse:
    """
    Process Linear webhook event.

    This webhook currently only listens for Issue events, and it treats both
    "Create" and "Update" events in the same way. It will set the identifier,
    team_name, title, state and estimate fields. The project_name and
    milestone_name are not included in the webhook payload.

    """
    try:
        body = json.loads(request.body.decode("utf-8"))
        data = body["data"]
        if body["type"] != "Issue":
            return HttpResponse("We are not interested in non-Issue updates")
    except (json.JSONDecodeError, KeyError):
        logger.exception("Unable to process Linear webhook event")
        return HttpResponse(
            "We couldn't process the request, but we're sending back a 200 anyway."
        )

    # we always get id, team, state in the payload. project/milestone are not included.
    id = data["id"]
    title = data["title"]
    team_name = data["team"]["name"]
    state = data["state"]["name"]
    estimate = data.get("estimate")
    identifier = f'{data["team"]["key"]}-{data["number"]}'

    try:
        issue = LinearIssue.objects.get(id=id)
    except LinearIssue.DoesNotExist:
        issue = LinearIssue.objects.create(id=id)

    issue.title = title
    issue.state = state
    issue.estimate = estimate
    issue.team_name = team_name
    issue.identifier = identifier
    issue.last_refreshed_at = tz_now()
    issue.save()
    return HttpResponse("Task updated")



from __future__ import annotations

from typing import Any

from django.conf import settings
from django.db import models


class TShirtSizing(models.IntegerChoices):
    XS = 1, "👕"
    S = 2, "👕👕"
    M = 3, "👕👕👕"
    L = 5, "👕👕👕👕"
    XL = 8, "👕👕👕👕👕"


class LinearIssue(models.Model):
    """Representation of a Linear issue."""

    id = models.UUIDField(primary_key=True)
    identifier = models.CharField(max_length=10)
    team_name = models.CharField("team", max_length=100)
    project_name = models.CharField("project", max_length=100, blank=True, default="")
    milestone_name = models.CharField(
        "milestone", max_length=100, blank=True, default=""
    )
    title = models.CharField(max_length=100, blank=True, default="")
    estimate = models.IntegerField(choices=TShirtSizing.choices, null=True)
    assignee_name = models.CharField("assigned to", max_length=100)
    state = models.CharField(max_length=100)
    created_at = models.DateTimeField(help_text="When this issue was created in Linear")
    last_updated_at = models.DateTimeField(
        help_text="When this issue was updated in Linear"
    )
    last_refreshed_at = models.DateTimeField(
        help_text=(
            "When this issue was last refreshed from Linear, via import or webhook"
        ),
    )

    class Meta:
        ordering = ["identifier"]
        permissions = [
            ("use_api", "Can run Linear API queries"),
        ]

    def __str__(self) -> str:
        return f"{self.identifier} - {self.title}"

    def get_absolute_url(self) -> str:
        """Return the Linear issue URL."""
        workspace = getattr(settings, "LINEAR_WORKSPACE_NAME")
        return f"https://linear.app/{workspace}/issue/{self.identifier}"

    def save(self, *args: Any, **kwargs: Any) -> LinearIssue:
        super().save(*args, **kwargs)
        return self

    def last_updated(self) -> str:
        """Return humanised version of the last_updated_at timestamp."""
        from django.contrib.humanize.templatetags import humanize

        return humanize.naturalday(self.last_updated_at)



from typing import List

from django.core.management.base import BaseCommand

from ...models import LinearIssue
from ...queries import pull_issues


class Command(BaseCommand):
    help = "Import all issues from Linear"

    def handle(self, *args, **options):
        issues: List[LinearIssue] = []
        self.stdout.write("Fetching first page")
        _issues, page_info = pull_issues()
        issues += _issues
        while page_info["hasNextPage"]:
            after = page_info["endCursor"]
            self.stdout.write(f"Fetching next page: {after}")
            _issues, page_info = pull_issues(after=after, save=True)
            issues += _issues
        self.stdout.write(f"Total issues: {len(issues)}")
        for issue in issues:
            self.stdout.write(str(issue))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['linear', 'linear.management.commands', 'linear.migrations']

package_data = \
{'': ['*'], 'linear': ['templates/admin/linear/*']}

install_requires = \
['django>=3.0,<4.0', 'requests']

setup_kwargs = {
    'name': 'django-linear',
    'version': '0.1.0',
    'description': 'Django app for importing Linear issues into the Django admin site.',
    'long_description': '# Django Linear\n\nReadonly access to Linear issues for users without a Linear account.\n\nThis app is designed to enable \'readonly\' user access to Linear issues via the Django admin site.\n\n### Motivation\n\nWe replaced our use of Jira with Linear a while back, and haven\'t looked back. However, one thing\nthat would improve it for our use case (small dev / design team, with many engaged internal\nstakeholders) would be the ability to share the status of issues more widely through the company.\n\nThis doesn\'t need to be sophisticated, we don\'t need to accept edits / comments - it\'s just a status\nupdate - for any issue, who is working on it, what its status is, which cycle / project it is part\nof.\n\n### Why bother?\n\nThe existing Google Sheets integration is almost good enough - but having the data in Django means\nwe can build additional functionality on top (custom notifications, alerts, reporting, etc.)\n\nThat said - this project is mainly an excuse to explore the GraphQL API.\n\n### Approach\n\nWe have a large "backoffice" project written in Django, and all our internal staff have accounts set\nup, and know their way around the Django admin site. Linear has a GraphQL API. Putting these two\ntogether, it ought to be simple to sync Linear updates to a Django model, and to surface those via\nthe admin site.\n\n![Screenshot of admin site](https://raw.githubusercontent.com/yunojuno/django-linear/master/screenshots/issue-list-view.png)\n\n### Principles\n\n1. Access is managed via existing Django authentication\n1. Data is readonly - if someone needs to edit an issue they should use Linear\n1. Users who require access to Linear should have a full ($) account\n\nNB this is **not** a tool to bypass Linear fees. Please respect their hard work.\n\n### How it works\n\nThe integration between Linear and the app occurs in two ways - via bulk import, and via webhook.\nThe data is readonly, so the principal is that all issues are imported once from Linear, and then\nmaintained via the webhook whenever they are updated. The webhook handler will pick up new Issues\ncreated after the import.\n\nThe integration doesn\'t go too deep into the data - we store the basics only - this is only intended\nfor use as simple dashboard for people who don\'t have / need access to Linear itself.\n\nYou can import the issues via the admin site itself (there is no "Add Linear issue" button), or if\nyou wish you can run the `import_issues` management command. If you don\'t have too many issues you\ncould even run the import on a schedule - start afresh each day.\n\n### Configuration\n\nIt\'s a standard Django app, so add it to `INSTALLED_APPS` as you would any other. The webhook URL is\n`/webook/`, so the recommended configuration is to add it to your main `urls.py` like this, making\nthe full url `/linear/webhook/`:\n\n```python\nurlpatterns = [\n    path("linear/", include("linear.urls")),\n]\n```\n\nYou should use this URL (with your fully-qualified domain name) when adding the webhook to Linear.\n\n#### Settings\n\nThe following Django settings are available:\n\n`LINEAR_API_KEY`\n\nThe only setting that is required is a valid API key, which is available from the Linear app, under\n"Workspace settings" > "API" > "Personal API keys".\n\n`LINEAR_API_PAGE_SIZE`\n\nThe page size to use when importing issues - defaults to 100, the max allowed by the API is 250.\n\n`LINEAR_WORKSPACE_NAME`\n\nYour workspace name is optional, but it is used in the admin site to provide a link from the object\npage to Linear - overriding the Django "View on site" link.\n',
    'author': 'YunoJuno',
    'author_email': 'code@yunojuno.com',
    'maintainer': 'YunoJuno',
    'maintainer_email': 'code@yunojuno.com',
    'url': 'https://github.com/yunojuno/django-linear',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from django.contrib import admin, messages
from django.db.models.query import QuerySet
from django.http.request import HttpRequest

from linear.queries import fetch_issue

from .models import LinearIssue


@admin.register(LinearIssue)
class LinearIssueAdmin(admin.ModelAdmin):
    list_display = (
        "identifier",
        "title",
        "milestone_name",
        "assignee_name",
        "estimate",
        "state",
    )
    list_filter = (
        "team_name",
        "state",
        "milestone_name",
        "assignee_name",
        "estimate",
        "created_at",
        "last_updated_at",
        "project_name",
    )
    search_fields = ("project_name", "identifier", "title")
    readonly_fields = (
        "id",
        "team_name",
        "project_name",
        "identifier",
        "title",
        "milestone_name",
        "assignee_name",
        "estimate",
        "state",
        "created_at",
        "last_updated_at",
        "last_refreshed_at",
    )
    actions = ("sync_issue",)

    def sync_issue(self, request: HttpRequest, queryset: QuerySet) -> None:
        count = queryset.count()
        for issue in queryset[:10]:
            update = fetch_issue(str(issue.id))
            self.message_user(
                request,
                f"Issue {update.identifier} was successfully updated.",
                messages.SUCCESS,
            )
        if count > 10:
            self.message_user(
                request,
                "Update ignored for remaining issues - use import command instead.",
                messages.WARNING,
            )
            return

    sync_issue.short_description = (  # type: ignore
        "Update selected linear issues via API"
    )



# Generated by Django 3.1.4 on 2020-12-16 17:56

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="LinearIssue",
            fields=[
                ("id", models.UUIDField(primary_key=True, serialize=False)),
                ("identifier", models.CharField(max_length=10)),
                ("team_name", models.CharField(max_length=100, verbose_name="team")),
                (
                    "project_name",
                    models.CharField(
                        blank=True, default="", max_length=100, verbose_name="project"
                    ),
                ),
                (
                    "milestone_name",
                    models.CharField(
                        blank=True, default="", max_length=100, verbose_name="milestone"
                    ),
                ),
                ("title", models.CharField(blank=True, default="", max_length=100)),
                (
                    "estimate",
                    models.IntegerField(
                        choices=[
                            (1, "👕"),
                            (2, "👕👕"),
                            (3, "👕👕👕"),
                            (5, "👕👕👕👕"),
                            (8, "👕👕👕👕👕"),
                        ],
                        null=True,
                    ),
                ),
                (
                    "assignee_name",
                    models.CharField(max_length=100, verbose_name="assigned to"),
                ),
                ("state", models.CharField(max_length=100)),
                (
                    "created_at",
                    models.DateTimeField(
                        help_text="When this issue was created in Linear"
                    ),
                ),
                (
                    "last_updated_at",
                    models.DateTimeField(
                        help_text="When this issue was updated in Linear"
                    ),
                ),
                (
                    "last_refreshed_at",
                    models.DateTimeField(
                        help_text="When this issue was last refreshed from Linear, via import or webhook"
                    ),
                ),
            ],
            options={
                "ordering": ["identifier"],
                "permissions": [("use_api", "Can run Linear API queries")],
            },
        ),
    ]
